import pandas as pd

def get_low(ids):
    low = None
    ret = None
    for id in ids:
        if low == None or int(id[3:]) < low:
            low = int(id[3:])
            ret = id
    return ret

def make_alias_dict(compound_aliases):
    compound_alias = {}
    for row_id, d in compound_aliases.iterrows():
        #print(row_id, d)
        alias = d['External ID']
        seed_id = d['ModelSEED ID']
        #print()
        if not alias in compound_alias:
            compound_alias[alias] = {'seed_id' : seed_id}
    return compound_alias

def get_lhs_rhs(seed_reaction, eval_func):
    result = {}
    result_no_zero = {}
    stoichiometry = seed_reaction['stoichiometry']
    if type(stoichiometry) == str:
        for reagent_str in stoichiometry.split(';'):
            reagent_data = reagent_str.split(':')
            cpd_id = reagent_data[1]
            value = float(reagent_data[0])
            if eval_func(value):
                if not cpd_id in result:
                    result[cpd_id] = 0
                result[reagent_data[1]] += value
            elif value == 0:
                raise Exception('zero value stoich: ' + stoichiometry)
        
    for cpd_id in result:
        if result[cpd_id] == 0:
            print(f'{seed_reaction["id"]} contains compound bein transported {cpd_id}')
        else:
            result_no_zero[cpd_id] = result[cpd_id]
    return result_no_zero

def get_stoichiometry(seed_reaction):
    result = {}
    result_no_zero = {}
    stoichiometry = seed_reaction['stoichiometry']
    if type(stoichiometry) == str:
        for reagent_str in stoichiometry.split(';'):
            reagent_data = reagent_str.split(':')
            cpd_id = reagent_data[1]
            if not cpd_id in result:
                result[cpd_id] = 0
            result[reagent_data[1]] += float(reagent_data[0])
        
    for cpd_id in result:
        if result[cpd_id] == 0:
            print(f'{seed_reaction["id"]} contains compound bein transported {cpd_id}')
        else:
            result_no_zero[cpd_id] = result[cpd_id]
    return result_no_zero
                  
def get_cstoichiometry(seed_reaction):
    result = {}
    result_no_zero = {}
    stoichiometry = seed_reaction['stoichiometry']
    #print(stoichiometry)
    if type(stoichiometry) == str:
        for reagent_str in stoichiometry.split(';'):
            reagent_data = reagent_str.split(':')
            value = reagent_data[0]
            cpd_id = reagent_data[1]
            cmp_index = reagent_data[2]
            if not (cpd_id, cmp_index) in result:
                result[(cpd_id, cmp_index)] = 0
            result[(cpd_id, cmp_index)] += float(value)
        
    for p in result:
        if result[p] == 0:
            print(f'{seed_reaction["id"]} contains compound bein transported {p}')
        else:
            result_no_zero[p] = result[p]
    return result_no_zero

class ModelSEEDObject:
    
    def __init__(self, data, api=None):
        self.data = data
        self.api = api
                  
    @property
    def id(self):
        return self.data['id']
                  
    @property
    def name(self):
        return self.data['name']
                  
class ModelSEEDCompound(ModelSEEDObject):
    
    @property
    def formula(self):
        return self.data['formula']
                  
    @property
    def database(self):
        return "seed.compound"

    @property
    def inchi(self):
        if not self.api == None:
            if self.id in self.api.compound_structures and 'InChI' in self.api.compound_structures[self.id]:
                  return self.api.compound_structures[self.id]['InChI']
        return None
                  
    @property
    def inchikey(self):
        if not self.api == None:
            if self.id in self.api.compound_structures and 'InChIKey' in self.api.compound_structures[self.id]:
                  return self.api.compound_structures[self.id]['InChIKey']
        if pd.isna(self.data['inchikey']):
            return None
        return self.data['inchikey']
                  
    @property
    def smiles(self):
        if not self.api == None:
            if self.id in self.api.compound_structures and 'SMILE' in self.api.compound_structures[self.id]:
                  return self.api.compound_structures[self.id]['SMILE']
        if pd.isna(self.data['smiles']):
            return None
        return self.data['smiles']
                  
    @property
    def aliases(self):
        if not self.api == None:
            if self.id in self.api.compound_aliases:
                  return self.api.compound_aliases[self.id]
        return {}
    
    @property
    def deltag(self):
        return self.data['deltag']
    
    @property
    def is_obsolete(self):
        if 'is_obsolete' in self.data:
            is_obsolete = self.data['is_obsolete']
            if is_obsolete == 0:
                return False
            else:
                return True
        return False
                  
class ModelSEEDReaction:
    
    def __init__(self, data, api=None):
        self.data = data
        self.api = api
        self.compounds = {}
        
    @property
    def id(self):
        return self.data['id']
    
    @property
    def stoichiometry(self):
        return get_stoichiometry(self.data)
                  
    @property
    def cstoichiometry(self):
        return get_cstoichiometry(self.data)

    @property
    def ec_numbers(self):
        if not self.api == None:
            if self.id in self.api.reaction_ecs:
                  return self.api.reaction_ecs[self.id]
        return {}
                  
    @property
    def database(self):
        return "seed.reaction"
                  
    @property
    def lhs(self):
        return get_lhs_rhs(self.data, lambda x : x < 0)
                  
    @property
    def rhs(self):
        return get_lhs_rhs(self.data, lambda x : x > 0)
    
    @property
    def is_transport(self):
        if 'is_transport' in self.data:
            is_transport = self.data['is_transport']
            if is_transport == 1:
                return True
                  
        if not len(self.lhs.keys() & self.rhs.keys()) == 0:
            return True
                  
        return False
                  
    @property
    def aliases(self):
        if not self.api == None:
            if self.id in self.api.reaction_aliases:
                  return self.api.reaction_aliases[self.id]
        return {}
    
    @property
    def is_obsolete(self):
        if 'is_obsolete' in self.data:
            is_obsolete = self.data['is_obsolete']
            if is_obsolete == 0:
                return False
            else:
                return True
        return False

class ModelSEED:
    
    def __init__(self, compounds, reactions, 
                 compound_aliases = {},
                 reaction_aliases = {},
                 compound_structures = {},
                 reaction_ecs = {}):
        self.compounds = compounds
        self.reactions = reactions
        self.compound_aliases = compound_aliases
        self.reaction_aliases = reaction_aliases
        self.compound_structures = compound_structures
        self.reaction_ecs = reaction_ecs
                  
    def summary(self):
        print("cpds:", len(self.compounds), "rxns:", len(self.reactions))
        print('structures', len(self.compound_structures))
        
    def get_formula(self, seed_id):
        return self.get_attribute('formula', seed_id)
    
    def get_name(self, seed_id):
        return self.get_attribute('name', seed_id)

    def get_attribute(self, attr, seed_id):
        seed = self.get_seed_compound(seed_id)
        if seed == None:
            return None
        if attr in seed:
            value = seed[attr]
            if not value == 'null':
                return seed[attr]
        return None
    
    def get_non_obsolete(self, seed_reaction):
        if not seed_reaction['is_obsolete'] == 0 and 'linked_reaction' in seed_reaction:
            for id in seed_reaction['linked_reaction'].split(';'):
                other = self.get_seed_reaction(id)
                if not other == None:
                    if other['is_obsolete'] == 0:
                        return other['id']

        return seed_reaction['id']

    def is_obsolete(self, seed_id):
        seed = self.get_seed_compound(seed_id)
        if seed == None:
            return None
        
        if 'is_obsolete' in seed:
            is_obsolete = seed['is_obsolete']
            if is_obsolete == 0:
                return False
            else:
                return True
        return None

    def get_seed_compound(self, seed_id):
        if seed_id in self.compounds:
            return ModelSEEDCompound(self.compounds[seed_id], self)
        return None
    
    def get_seed_reaction(self, seed_id):
        if seed_id in self.reactions:
            return ModelSEEDReaction(self.reactions[seed_id], self)
        return None

    def is_abstract(self, seed_id):
        return None
    
    def get_seed_compound_by_alias(self, database, cpd_id):
        o = None
        for o_id in self.compounds:
            aliases_str = self.compounds[o_id]['aliases']
            alias = dict((a.split(':')[0], a.split(':')[1]) for a in aliases_str.split(';'))
            if database in alias and cpd_id in alias[database].split('|'):
                o = self.compounds[o_id]
                break
        return o
    
    def get_seed_reaction_by_alias(self, id):
        o = None
        for o_id in self.reactions:
            if id in self.reactions[o_id]['aliases']:
                o = self.reactions[o_id]
                break
        return o

def get_aliases_from_df(df):
    aliases = {}
    for t in df.itertuples():
        seed_id = t[1]
        database_id = t[2]
        database = t[3]
        if not seed_id in aliases:
            aliases[seed_id] = {}
        if not database in aliases[seed_id]:
            aliases[seed_id][database] = set()
        aliases[seed_id][database].add(database_id)
    return aliases
    
def from_github(commit):
    print(commit)
    #https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase/566a42f8c04a0dab4536b59377074ad23bcc08be/Biochemistry/reactions.tsv
    database_repo = 'https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase'
    reactions_url = database_repo + '/%s/Biochemistry/reactions.tsv' % commit
    compounds_url = database_repo + '/%s/Biochemistry/compounds.tsv' % commit
    reactions_aliases_url = database_repo + '/%s/Biochemistry/Aliases/Unique_ModelSEED_Reaction_Aliases.txt' % commit
    compounds_aliases_url = database_repo + '/%s/Biochemistry/Aliases/Unique_ModelSEED_Compound_Aliases.txt' % commit
    compounds_structures_url = database_repo + '/%s/Biochemistry/Structures/Unique_ModelSEED_Structures.txt' % commit
    reactions_ec_url = database_repo + '/%s/Biochemistry/Aliases/Unique_ModelSEED_Reaction_ECs.txt' % commit
    compounds = {}
    reactions = {}

    print('load:', reactions_url)
    seed_reactions = pd.read_csv(reactions_url, sep='\t', low_memory=False)
    
    for row_id, d in seed_reactions.iterrows():
        seed_reaction = build_reaction(d)
        reactions[seed_reaction['id']] = seed_reaction
        
    print('load:', compounds_url)
    seed_compounds = pd.read_csv(compounds_url, sep='\t', low_memory=False) #Columns (10,14,15) have mixed types.
    
    for row_id, d in seed_compounds.iterrows():
        seed_compound = build_compound(d)
        compounds[seed_compound['id']] = seed_compound
        
        
    compound_aliases = pd.read_csv(compounds_aliases_url, sep='\t')
    reaction_aliases = pd.read_csv(reactions_aliases_url, sep='\t')
    
    alias_to_database_to_cpd = get_aliases_from_df(compound_aliases)
    alias_to_database_to_rxn = get_aliases_from_df(reaction_aliases)
    print('load:', reactions_ec_url)
    df_reaction_ecs = pd.read_csv(reactions_ec_url, sep='\t')
    print('load:', compounds_structures_url)
    df_structures = pd.read_csv(compounds_structures_url, sep='\t')

    compound_structures = get_structures(df_structures)
    reaction_ecs = get_aliases_from_df(df_reaction_ecs)
    
    modelseed = ModelSEED(compounds, reactions, alias_to_database_to_cpd, alias_to_database_to_rxn, compound_structures, reaction_ecs)
    return modelseed

def from_local(path):
    database_repo = path
    reactions_url = database_repo + '/Biochemistry/reactions.tsv'
    compounds_url = database_repo + '/Biochemistry/compounds.tsv'
    reactions_aliases_url = database_repo + '/Biochemistry/Aliases/Unique_ModelSEED_Reaction_Aliases.txt' 
    compounds_aliases_url = database_repo + '/Biochemistry/Aliases/Unique_ModelSEED_Compound_Aliases.txt' 
    reactions_names_url = database_repo + '/Biochemistry/Aliases/Unique_ModelSEED_Reaction_Names.txt'
    compounds_names_url = database_repo + '/Biochemistry/Aliases/Unique_ModelSEED_Compound_Names.txt'
    compounds_structures_url = database_repo + '/Biochemistry/Structures/Unique_ModelSEED_Structures.txt'
    reactions_ec_url = database_repo + '/Biochemistry/Aliases/Unique_ModelSEED_Reaction_ECs.txt'
    
    compounds = {}
    reactions = {}
    
    print('load:', reactions_url)
    seed_reactions = pd.read_csv(reactions_url, sep='\t', low_memory=False)
    print('load:', compounds_url)
    seed_compounds = pd.read_csv(compounds_url, sep='\t', low_memory=False) #Columns (10,14,15) have mixed types.
    
    for row_id, d in seed_reactions.iterrows():
        seed_reaction = build_reaction(d)
        reactions[seed_reaction['id']] = seed_reaction

    for row_id, d in seed_compounds.iterrows():
        seed_compound = build_compound(d)
        compounds[seed_compound['id']] = seed_compound
        
    print('load:', compounds_structures_url)
    df_structures = pd.read_csv(compounds_structures_url, sep='\t')
    compound_structures = get_structures(df_structures)

    print('load:', compounds_aliases_url)
    df_compound_aliases = pd.read_csv(compounds_aliases_url, sep='\t')
    print('load:', reactions_aliases_url)
    df_reaction_aliases = pd.read_csv(reactions_aliases_url, sep='\t')
    print('load:', reactions_ec_url)
    df_reaction_ecs = pd.read_csv(reactions_ec_url, sep='\t')
    
    compound_aliases = get_aliases_from_df(df_compound_aliases)
    reaction_aliases = get_aliases_from_df(df_reaction_aliases)
    reaction_ecs = get_aliases_from_df(df_reaction_ecs)
    
    modelseed = ModelSEED(compounds, reactions, compound_aliases, reaction_aliases, compound_structures, reaction_ecs)
    
    return modelseed

def get_structures(seed_structures):
    compound_structures = {}
    
    for row_id, d in seed_structures.iterrows():
        seed_id = d['ID']
        t = d['Type']
        value = d['Structure']
        if not seed_id in compound_structures:
            compound_structures[seed_id] = {}
        if t in compound_structures[seed_id]:
            print('warning duplicate structure:', t, seed_id)
        compound_structures[seed_id][t] = value
    
    return compound_structures
                  
def build_compound(d):
    seed_compound = {
        'id': d['id'],
        'abbreviation': d['abbreviation'],
        'name': d['name'],
        'formula': d['formula'],
        'mass': d['mass'],
        'source': d['source'],
        'inchikey': d['inchikey'],
        'charge': d['charge'],
        'is_core': d['is_core'],
        'is_obsolete': d['is_obsolete'],
        'linked_compound': d['linked_compound'],
        'is_cofactor': d['is_cofactor'],
        'deltag': d['deltag'],
        'deltagerr': d['deltagerr'],
        'pka': d['pka'],
        'pkb': d['pkb'],
        'abstract_compound': d['abstract_compound'],
        'comprised_of': d['comprised_of'],
        'aliases': d['aliases'],
        'smiles': d['smiles']
    }
    return seed_compound

def build_reaction(d):
    seed_reaction = {
            'id': d['id'],
            'abbreviation': d['abbreviation'],
            'name': d['name'],
            'code': d['code'],
            'stoichiometry': d['stoichiometry'],
            'is_transport': d['is_transport'],
            'equation': d['equation'],
            'definition': d['definition'],
            'reversibility': d['reversibility'],
            'direction': d['direction'],
            'abstract_reaction': d['abstract_reaction'],
            'pathways': d['pathways'],
            'aliases': d['aliases'],
            'ec_numbers': d['ec_numbers'],
            'deltag': d['deltag'],
            'deltagerr': d['deltagerr'],
            'compound_ids': d['compound_ids'],
            'status': d['status'],
            'is_obsolete': d['is_obsolete'],
            'linked_reaction': d['linked_reaction'],
            'notes': d['notes'],
            'source': d['source']
    }
    return seed_reaction


import logging
import math

logger = logging.getLogger(__name__)

def multiply(s, v):
    s_ = {}
    for i in s:
        s_[i] = v * s[i]
    return s_

def to_faa(kgenome):
    faa_features = []
    
    for feature in kgenome['features']:
        faa_features.append('>' + feature['id'] + '\n' + feature['protein_translation'])
        #print(feature)
        #break
    
    return '\n'.join(faa_features)

def get_lower(s):
    return sorted(list(s))[0]

def get_locked(compartment_config):
    #print('get_locked', compartment_config)
    locked = {}
    for socket in compartment_config:
        if len(compartment_config[socket]) == 1:
            locked[socket] = compartment_config[socket].pop()
    return locked

def get_compartment_config(model_reaction, seed_reaction, locked = {}, depth = 0):
    model_reaction_reagents = model_reaction['modelReactionReagents']
    stoichiometry = seed_reaction['stoichiometry']
    #print('locked', locked)
    #print(stoichiometry)
    #print(model_reaction)
    
    compartment_config = {}
    for socket in locked:
        compartment_config[socket] = set([locked[socket]])
    all_compartments = set()
    #print(depth, 'initial', compartment_config)
    for p in stoichiometry.split(';'):
        o = p.split(':')
        #cmp = o[]
        value = float(o[0])
        compound_id = o[1]
        compartment_socket = int(o[2])
        if not compartment_socket in locked:
            if not compartment_socket in compartment_config:
                compartment_config[compartment_socket] = set()
            #print(depth, 'match socket', compartment_socket)
            for o in model_reaction_reagents:
                if compound_id in o['modelcompound_ref']:
                    cmp = o['modelcompound_ref'].split('_')[-1]
                    #print(depth, cmp, locked.values())
                    if not cmp in locked.values():
                        compartment_config[compartment_socket].add(cmp)
                    all_compartments.add(cmp)
    
    #print(depth, 'match result', compartment_config)

    
    all_compartments = sorted(list(all_compartments))
    
    if len(all_compartments) > 1:
        for socket in compartment_config:
            if len(compartment_config[socket]) > 1:
                compartment_config[socket] = set([get_lower(compartment_config[socket])])
                break
    
    l = get_locked(compartment_config)
    
    #print(depth, 'locked', l)
    
    if depth > 10:
        print('unable to detect valid configuration')
        return None
    if not l == locked:
        return get_compartment_config(model_reaction, seed_reaction, locked = l, depth = depth + 1)
    
    compartment_config = locked
    #print('yes!', compartment_config)
    
    for socket in compartment_config:
        if len(compartment_config[socket]) == 1:
            compartment_config[socket] = compartment_config[socket].pop()
    
    valid = True
    for socket in compartment_config:
        if not isinstance(compartment_config[socket], str):
            valid = False
            
    #print('valid', valid)
    if valid:
        return compartment_config
    
    return None



def get_reaction_compartment2(compartment_config):
    cmps = set(compartment_config.values())
    if not cmps == None:
        if len(cmps) == 1:
            return cmps.pop()
        elif 'c0' in cmps and len(cmps) == 2:
            cmps.remove('c0')
            return cmps.pop()
        print('fail', cmps)
    else:
        return None
    return "z0"

def get_bounds(r):
    maxrevflux = r['maxrevflux']
    maxforflux = r['maxforflux']
    direction = '='
    if maxrevflux == 0 and maxforflux > 0:
        direction = '>'
    elif maxrevflux > 0 and maxforflux == 0:
        direction = '<'
    elif maxrevflux == 0 and maxforflux == 0:
        direction = '0'
        
    return maxrevflux, maxforflux, direction

def seed_coefficients_to_string(coeff_list, op = '<=>'):
    lhs = []
    rhs = []
    pair_to_str = lambda v, id : id if math.fabs(v) == 1 else str(math.fabs(v)) + ' ' + id
    for o in coeff_list:
        cpd_id = o['modelcompound_ref'].split('/')[-1]
        value = o['coefficient']
        if value > 0:
            rhs.append(pair_to_str(value, cpd_id))
        elif value < 0:
            lhs.append(pair_to_str(value, cpd_id))
        else:
            logger.warning("zero value found: %s", o)
    return "{} {} {}".format(' + '.join(lhs), op, ' + '.join(rhs))

def print_stoich(stoich, eq="<=>"):
    l = {}
    r = {}
    for o in stoich:
        v = stoich[o]
        v_text = ""
        if not 1 == math.fabs(v):
            v_text = str(math.fabs(v)) + " "
        if v < 0:
            l[o] = v_text
        elif v > 0:
            r[o] = v_text
        else:
            print('warn: found zero value stoich', stoich)
            
    l_text = []
    r_text = []
    for i in l:
        l_text.append(l[i] + i)
    for i in r:
        r_text.append(r[i] + i)
    #print(l_text, r_text)
    return ' + '.join(l_text) + " " + eq + " " + ' + '.join(r_text)

def get_str(k, def_value, d):
    if k in d:
        if d[k] == None:
            return def_value
        return str(d[k])
    return def_value

def get_int(k, def_value, d):
    if k in d:
        if d[k] == None:
            return def_value
        return int(d[k])
    return def_value

def get_id_from_ref(str, stok='/'):
    return str.split(stok)[-1]


from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref

class KBaseGenome(KBaseObjectBase):
    
    def read_genome_aliases(self):
        gene_aliases = {}
        for feature in self.data['features']:
            gene_aliases[feature['id']] = {}
            if 'aliases' in feature and type(feature['aliases']) == list:
                for alias in feature['aliases']:
                    if type(alias) == list:
                        for a in alias:
                            #risk parsing the first element
                            self.get_old_alias(a, feature, gene_aliases)
                    else:
                        self.get_old_alias(alias, feature, gene_aliases)
                        #print('discard', alias)
            #print(feature['aliases' in ])
            db_refs = self.get_db_xrefs(feature)
            gene_aliases[feature['id']].update(db_refs)
        return gene_aliases
    
    def get_db_xrefs(self, feature):
        aliases = {}
        if 'db_xrefs' in feature:
            for ref in feature['db_xrefs']:
                db = ref[0]
                value = ref[1]
                if db == 'EcoGene':
                    aliases['ecogene'] = value
                elif db == 'GeneID':
                    aliases['ncbigene'] = value
        return aliases
    
    def get_old_alias(self, alias, feature, gene_aliases):
        if alias.startswith('EcoGene:'):
            gene_aliases[feature['id']]['ecogene'] = alias.split(':')[1]
        elif alias.startswith('UniProtKB/Swiss-Prot:'):
            gene_aliases[feature['id']]['uniprot'] = alias.split(':')[1]
        elif alias.startswith('NP_'):
            gene_aliases[feature['id']]['ncbiprotein'] = alias
        elif alias.startswith('ASAP:'):
            gene_aliases[feature['id']]['asap'] = alias.split(':')[1]
        elif alias.startswith('GI:'):
            gene_aliases[feature['id']]['ncbigi'] = alias
        elif alias.startswith('GeneID:'):
            gene_aliases[feature['id']]['ncbigene'] = alias.split(':')[1]
        else:
            1


from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref

class KBaseFBA(KBaseObjectBase):
    
    def get_reaction_variable_by_id(self, rxn_id):
        for v in self.dara['FBAReactionVariables']:
            v_id = v['modelreaction_ref'].split('/')[-1]
            if rxn_id == v_id:
                return v
        return None
    
    @property
    def objective_value(self):
        return self.data['objectiveValue']


import re

class KBaseFBAModelBuilder:
    
    def __init__(self, fbamodel, modelseed):
        self.fbamodel = fbamodel
        self.modelseed = modelseed

def get_mapping(df, valid_databases, pd):
    mapping = {}
    for id, row in df.iterrows():
        mapping[id] = {}
        #print(dir(row))
        for db in df.keys():
            #print(db)
            if db in valid_databases:
                value = row[db]
                if not pd.isna(value):
                    mapping[id][db] = value
    return mapping

def atom_count(formula):
    atoms = {}
    pairs = re.findall(r'([A-Z][a-z]*)(\d*)', formula)
    for p in pairs:
        if not p[0] in atoms:
            atoms[p[0]] = 0
        v = p[1]
        if len(v) == 0:
            v = 1
        atoms[p[0]] += int(v)
        #print(p[0], v)
    return atoms

def is_valid_formula(f, pt):
    atoms = atom_count(f)
    for e in atoms:
        name = pt.get_element_name(e)
        if name == None:
            return False
    return True

class PeriodicTable:
    
    def __init__(self):
        self.elements_v = {
            "Ac": 89,
            "Ag": 47,
            "Al": 13,
            "Am": 95,
            "Ar": 18,
            "As": 33,
            "At": 85,
            "Au": 79,
            "B": 5,
            "Ba": 56,
            "Be": 4,
            "Bh": 107,
            "Bi": 83,
            "Bk": 97,
            "Br": 35,
            "C": 6,
            "Ca": 20,
            "Cd": 48,
            "Ce": 58,
            "Cf": 98,
            "Cl": 17,
            "Cm": 96,
            "Cn": 112,
            "Co": 27,
            "Cr": 24,
            "Cs": 55,
            "Cu": 29,
            "Db": 105,
            "Ds": 110,
            "Dy": 66,
            "Er": 68,
            "Es": 99,
            "Eu": 63,
            "F": 9,
            "Fe": 26,
            "Fl": 114,
            "Fm": 100,
            "Fr": 87,
            "Ga": 31,
            "Gd": 64,
            "Ge": 32,
            "H": 1,
            "He": 2,
            "Hf": 72,
            "Hg": 80,
            "Ho": 67,
            "Hs": 108,
            "I": 53,
            "In": 49,
            "Ir": 77,
            "K": 19,
            "Kr": 36,
            "La": 57,
            "Li": 3,
            "Lr": 103,
            "Lu": 71,
            "Lv": 116,
            "Mc": 115,
            "Md": 101,
            "Mg": 12,
            "Mn": 25,
            "Mo": 42,
            "Mt": 109,
            "N": 7,
            "Na": 11,
            "Nb": 41,
            "Nd": 60,
            "Ne": 10,
            "Nh": 113,
            "Ni": 28,
            "No": 102,
            "Np": 93,
            "O": 8,
            "Og": 118,
            "Os": 76,
            "P": 15,
            "Pa": 91,
            "Pb": 82,
            "Pd": 46,
            "Pm": 61,
            "Po": 84,
            "Pr": 59,
            "Pt": 78,
            "Pu": 94,
            "Ra": 88,
            "Rb": 37,
            "Re": 75,
            "Rf": 104,
            "Rg": 111,
            "Rh": 45,
            "Rn": 86,
            "Ru": 44,
            "S": 16,
            "Sb": 51,
            "Sc": 21,
            "Se": 34,
            "Sg": 106,
            "Si": 14,
            "Sm": 62,
            "Sn": 50,
            "Sr": 38,
            "Ta": 73,
            "Tb": 65,
            "Tc": 43,
            "Te": 52,
            "Th": 90,
            "Ti": 22,
            "Tl": 81,
            "Tm": 69,
            "Ts": 117,
            "U": 92,
            "V": 23,
            "W": 74,
            "Xe": 54,
            "Y": 39,
            "Yb": 70,
            "Zn": 30,
            "Zr": 40
        }
        self.elements = {
            "Ac" : "Actinium",
            "Ag" : "Silver",
            "Al" : "Aluminum",
            "Am" : "Americium",
            "Ar" : "Argon",
            "As" : "Arsenic",
            "At" : "Astatine",
            "Au" : "Gold",
            "B" : "Boron",
            "Ba" : "Barium",
            "Be" : "Beryllium",
            "Bh" : "Bohrium",
            "Bi" : "Bismuth",
            "Bk" : "Berkelium",
            "Br" : "Bromine",
            "C" : "Carbon",
            "Ca" : "Calcium",
            "Cd" : "Cadmium",
            "Ce" : "Cerium",
            "Cf" : "Californium",
            "Cl" : "Chlorine",
            "Cm" : "Curium",
            "Cn" : "Copernicium",
            "Co" : "Cobalt",
            "Cr" : "Chromium",
            "Cs" : "Cesium",
            "Cu" : "Copper",
            "Db" : "Dubnium",
            "Ds" : "Darmstadtium",
            "Dy" : "Dysprosium",
            "Er" : "Erbium",
            "Es" : "Einsteinium",
            "Eu" : "Europium",
            "F" : "Fluorine",
            "Fe" : "Iron",
            "Fl" : "Flerovium",
            "Fm" : "Fermium",
            "Fr" : "Francium",
            "Ga" : "Gallium",
            "Gd" : "Gadolinium",
            "Ge" : "Germanium",
            "H" : "Hydrogen",
            "He" : "Helium",
            "Hf" : "Hafnium",
            "Hg" : "Mercury",
            "Ho" : "Holmium",
            "Hs" : "Hassium",
            "I" : "Iodine",
            "In" : "Indium",
            "Ir" : "Iridium",
            "K" : "Potassium",
            "Kr" : "Krypton",
            "La" : "Lanthanum",
            "Li" : "Lithium",
            "Lr" : "Lawrencium",
            "Lu" : "Lutetium",
            "Lv" : "Livermorium",
            "Mc" : "Moscovium",
            "Md" : "Mendelevium",
            "Mg" : "Magnesium",
            "Mn" : "Manganese",
            "Mo" : "Molybdenum",
            "Mt" : "Meitnerium",
            "N" : "Nitrogen",
            "Na" : "Sodium",
            "Nb" : "Niobium",
            "Nd" : "Neodymium",
            "Ne" : "Neon",
            "Nh" : "Nihonium",
            "Ni" : "Nickel",
            "No" : "Nobelium",
            "Np" : "Neptunium",
            "O" : "Oxygen",
            "Og" : "Oganesson",
            "Os" : "Osmium",
            "P" : "Phosphorus",
            "Pa" : "Protactinium",
            "Pb" : "Lead",
            "Pd" : "Palladium",
            "Pm" : "Promethium",
            "Po" : "Polonium",
            "Pr" : "Praseodymium",
            "Pt" : "Platinum",
            "Pu" : "Plutonium",
            "Ra" : "Radium",
            "Rb" : "Rubidium",
            "Re" : "Rhenium",
            "Rf" : "Rutherfordium",
            "Rg" : "Roentgenium",
            "Rh" : "Rhodium",
            "Rn" : "Radon",
            "Ru" : "Ruthenium",
            "S" : "Sulfur",
            "Sb" : "Antimony",
            "Sc" : "Scandium",
            "Se" : "Selenium",
            "Sg" : "Seaborgium",
            "Si" : "Silicon",
            "Sm" : "Samarium",
            "Sn" : "Tin",
            "Sr" : "Strontium",
            "Ta" : "Tantalum",
            "Tb" : "Terbium",
            "Tc" : "Technetium",
            "Te" : "Tellurium",
            "Th" : "Thorium",
            "Ti" : "Titanium",
            "Tl" : "Thallium",
            "Tm" : "Thulium",
            "Ts" : "Tennessine",
            "U" : "Uranium",
            "V" : "Vanadium",
            "W" : "Tungsten",
            "Xe" : "Xenon",
            "Y" : "Yttrium",
            "Yb" : "Ytterbium",
            "Zn" : "Zinc",
            "Zr" : "Zirconium"
        }
        
    def get_element_name(self, e):
        if e in self.elements:
            return self.elements[e]
        return None



class ModelSeedBuilder:
    
    def __init__(self, fbamodel, modelseed):
        self.fbamodel = fbamodel
        self.modelseed = modelseed
        
    def configure_reaction(self, seed_reaction, direction, compartment_config, model_reaction_proteins):
        maxforflux = 1000
        maxrevflux = 1000
        if direction == '>':
            maxrevflux = 0
        elif direction == '<':
            maxforflux = 0

        if not seed_reaction['is_obsolete'] == 0:
            print('warning obsolete reaction', seed_id)

        model_reaction_reagents = configure_stoichiometry(
            seed_reaction, 
            compartment_config)

        compartment = cobrakbase.core.utils.get_reaction_compartment2(compartment_config)

        if len(compartment) > 1:
            compartment = compartment[0]

        modelreaction = {
            'aliases': [], 
            'dblinks': {}, 
            'direction': direction, 
            'edits': {}, 
            'gapfill_data': {}, 
            'id': "{}_{}".format(seed_reaction['id'], compartment),
            'maxforflux': maxforflux, 
            'maxrevflux': maxrevflux, 
            'modelReactionProteins': model_reaction_proteins, 
            'modelReactionReagents': model_reaction_reagents, 
            'modelcompartment_ref': '~/modelcompartments/id/c0', 
            'name': seed_reaction['name'], 
            'numerical_attributes': {}, 
            'probability': 0, 
            'protons': 0, 
            'reaction_ref': '~/template/reactions/id/{}_{}'.format(seed_reaction['id'], compartment), 
            'string_attributes': {}
        }
        
        return modelreaction


import logging
import os
import json
import cobra
from cobrakbase.core.utils import get_str, get_int, get_id_from_ref
from cobrakbase.core.converters import KBaseFBAModelToCobraBuilder
from cobrakbase.kbaseapi import KBaseAPI
from cobrakbase.Workspace.WorkspaceClient import Workspace as WorkspaceClient
from cobra.core import Gene, Metabolite, Model, Reaction
from cobra.util.solver import linear_reaction_coefficients
import cobrakbase.core.model
import cobrakbase.core.converters
import cobrakbase.modelseed.utils
import cobrakbase.modelseed.modelseed

__author__  = "Filipe Liu"
__email__   = "fliu@anl.gov"
__version__ = "0.1.9"

print("cobrakbase", __version__)

KBASE_WS_URL = "https://kbase.us/services/ws/"
DEV_KBASE_WS_URL = "https://appdev.kbase.us/services/ws/"

SBO_ANNOTATION = "sbo"

bigg = False

API = None

logging.basicConfig()
logger = logging.getLogger(__name__)

def read_model(id, workspace):
    if API == None:
        print("Login first! cobrakbase.login(<API_TOKEN>)")
        return None
    data = API.get_object(id, workspace)
    return convert_kmodel(data)

def read_model_with_media(modelId, mediaId, workspace):
    if API == None:
        print("Login first! cobrakbase.login(<API_TOKEN>)")
        return None
    model_data = API.get_object(modelId, workspace)
    media_data = API.get_object(mediaId, workspace)
    media = convert_media(media_data)
    #print(media)
    return convert_kmodel(model_data, media)

def login(token, dev=False):
    global API
    API = KBaseAPI(token, dev)

SINK = ["cpd02701_c0", "cpd11416_c0", "cpd15302_c0", "cpd11416_c0"]

COBRA_DEFAULT_LB = -1000000
COBRA_DEFAULT_UB =  1000000
COBRA_0_BOUND = 0

def convert_biomass_to_reaction(biomass, mets):
    mr_id = biomass['id'] + "_biomass"
    name = biomass['name']
    object_stoichiometry = {}
    for biomasscompound in biomass['biomasscompounds']:
        coefficient = biomasscompound['coefficient']
        modelcompound_ref = biomasscompound['modelcompound_ref']
        mc_id = get_id_from_ref(modelcompound_ref)
        met_id = build_cpd_id(mc_id)
        met = mets[mc_id]
        object_stoichiometry[met] = coefficient
    reaction = Reaction(id=build_rxn_id(mr_id), name=name, lower_bound=COBRA_0_BOUND, upper_bound=COBRA_DEFAULT_UB)
    objective_id = reaction.id
    reaction.add_metabolites(object_stoichiometry)
    #SBO:0000629 [biomass production]
    #Biomass production, often represented 'R_BIOMASS_', is usually the optimization target reaction of constraint-based models ...
    reaction.annotation[SBO_ANNOTATION] = "SBO:0000629" 
    return reaction

def get_reaction_constraints(data):
    #clean this function !
    if 'maxrevflux' in data and 'maxforflux' in data:
        if data['maxrevflux'] == 1000000 and data['maxforflux'] == 1000000:
            if 'direction' in data:
                if data['direction'] == '>':
                    return 0, 1000
                elif data['direction'] == '<':
                    return -1000, 0
                else:
                    return -1000, 1000
            else:
                return -1 * data['maxrevflux'], data['maxforflux']
        else:
            return -1 * data['maxrevflux'], data['maxforflux']
    return -1000, 1000

def convert_kmodel(kmodel, media=None, exchanges = True, model_id = "kbase"):
    model_test = cobra.Model(model_id)

    comps = {}
    mets = {}
    sink = set()
    reactions = []
    extra = set()

    for mcomp in kmodel['modelcompartments']:
        mcomp_id = mcomp['id']
        name = mcomp['label']
        comps[mcomp_id] = name

    for mc in kmodel["modelcompounds"]:
        #print(mc.keys())
        formula = None
        if not mc['formula'] == 'null':
            formula = mc['formula']
        name = mc['name']
        charge = get_int('charge', 0, mc)
        mc_id = mc['id']
        annotation = {}
        if 'dblinks' in mc:
            annotation = get_cpd_annotation(mc['dblinks'])
        compartment = get_compartment_id(mc, simple=True)
        id = build_cpd_id(mc_id)
        if bigg:
            if "bigg.metabolite" in annotation:
                id = annotation["bigg.metabolite"] + "_" + compartment
                #print(id)
        
        if mc_id in SINK:
            logger.info('Add Sink: [%s]', mc_id)
            extra.add(mc_id)
            sink.add(mc_id)
        if compartment.startswith("e"):
            extra.add(mc_id)
            
        met = Metabolite(id=id, formula=formula, name=name, charge=charge, compartment=compartment)
        met.annotation[SBO_ANNOTATION] = "SBO:0000247" #simple chemical - Simple, non-repetitive chemical entity.
        if id.startswith('cpd'):
            met.annotation["seed.compound"] = id.split("_")[0]
        #met.annotation[""] = "!!!"
        met.annotation.update(annotation)
        mets[mc_id] = met

    genes = set()
        #print(mc)
    for mr in kmodel["modelreactions"]:
        mr_id = mr['id']
        name = mr['name']
        
        lower_bound, upper_bound = get_reaction_constraints(mr)
        annotation = {}
        if 'dblinks' in mr:
            annotation = get_rxn_annotation(mr['dblinks'])
        id = build_rxn_id(mr_id)
        if bigg:
            if "bigg.reaction" in annotation:
                id = annotation["bigg.reaction"]
        #print(id)
        reaction = Reaction(id=id, name=name, lower_bound=lower_bound, upper_bound=upper_bound)
        #print(mr['maxrevflux'], mr['maxforflux'], reaction.lower_bound)
        reaction.annotation[SBO_ANNOTATION] = "!!!"
        if id.startswith('rxn'):
            reaction.annotation["seed.reaction"] = id.split("_")[0]
        reaction.annotation.update(annotation)
        object_stoichiometry = {}
        for mrr in mr['modelReactionReagents']:
            modelcompound_ref = mrr['modelcompound_ref']
            coefficient = mrr['coefficient']
            mc_id = get_id_from_ref(modelcompound_ref)
            met_id = build_cpd_id(mc_id)
            met = mets[mc_id]#model_test.metabolites.get_by_id(met_id)
            #print(met, met_id, coefficient)
            object_stoichiometry[met] = coefficient
        reaction.annotation[SBO_ANNOTATION] = "SBO:0000176" #biochemical reaction
        reaction.add_metabolites(object_stoichiometry)
        gpr = get_gpr(mr)
        gpr_string = get_gpr_string(gpr)
        #print(gpr_string)
        reaction.gene_reaction_rule = gpr_string
        genes |= get_genes(gpr)
        #print(reaction)
        reactions.append(reaction)
        #print(mr.keys())

    objective_id = None
    for biomass in kmodel['biomasses']:
        reaction = convert_biomass_to_reaction(biomass, mets)
        reactions.append(reaction)
        objective_id = reaction.id
        #print(biomass)
    #print(media)
    
    if exchanges:
        logger.info('Setup Drains. EX: %d SK: %d', len(extra), len(sink))
        for e in extra:
            met = mets[e]
            prefix = "EX_"
            if e in sink:
                prefix = "DM_"
            id = prefix + met.id
            lower_bound = COBRA_DEFAULT_LB
            upper_bound = COBRA_DEFAULT_UB
            if not media == None:
                lower_bound = 0
            if not media == None and e.split("_")[0] in media:
                ct = media[e.split("_")[0]]
                lower_bound = ct[0]
                upper_bound = ct[1]

            #print(e, met, id, lower_bound, upper_bound)

            object_stoichiometry = {met : -1}
            reaction = Reaction(id=id, name="Exchange for " + met.name, lower_bound=lower_bound, upper_bound=upper_bound)
            reaction.add_metabolites(object_stoichiometry)
            reaction.annotation[SBO_ANNOTATION] = "SBO:0000627" #exchange reaction - ... provide matter influx or efflux to a model, for example to replenish a metabolic network with raw materials ... 
            reactions.append(reaction)
            #print(reaction.name)
    #print("Genes:", genes)
    for g in genes:
        gene = Gene(id=build_gene_id(g), name=g)
        gene.annotation[SBO_ANNOTATION] = "SBO:0000243"
        model_test.genes.append(gene)

    model_test.compartments = comps
    #model_test.add_metabolites(mets.values)
    try:
        model_test.add_reactions(reactions)
    except ValueError as e:
        warn(str(e))
    
    if not objective_id == None:
        model_test.objective = model_test.reactions.get_by_id(id=objective_id)
        linear_reaction_coefficients(model_test)
    return model_test


def read_kbase_model(filename):
    kmodel = {}
    with open(filename) as file:
        data = file.read()
        kmodel = json.loads(data)
    return kmodel

def read_json(filename, converter):
    di = None
    if os.path.exists(filename):
        with open(filename) as file:
            data = file.read()
            di = json.loads(data)
    if di is None:
        return None
    return converter(di)

def convert_media(kmedia):
    media = {}
    for mediacompound in kmedia['mediacompounds']:
        met_id = get_id_from_ref(mediacompound['compound_ref'])
        lb = -1 * mediacompound['maxFlux']
        ub = -1 * mediacompound['minFlux']
        media[met_id] = (lb, ub)
    return media

def read_kbase_media(filename):
    return read_json(filename, convert_media)

def get_compartment_id(modelcompound, simple=False):
    modelcompartment_ref = modelcompound['modelcompartment_ref']
    compartment = get_id_from_ref(modelcompartment_ref)
    if simple:
        return compartment[0]
    return compartment

def get_cpd_annotation(dblinks):
    annotation = {}
    
    for db in dblinks:
        if db == "BiGG2":
            annotation["bigg.metabolite"] = dblinks[db][0]
        if db == "LigandCompound":
            annotation["kegg.compound"] = dblinks[db][0]
        if db == "ChEBI":
            annotation["chebi"] = dblinks[db][0]
        if db == "MetaNetX":
            annotation["metanetx.chemical"] = dblinks[db][0]
        if db == "MetaCyc":
            annotation["biocyc"] = dblinks[db][0]
        if db == "ModelSeed":
            annotation["seed.compound"] = dblinks[db][0]
        if db == "HMDB":  
            annotation["hmdb"] = dblinks[db][0]
    return annotation

def get_rxn_annotation(dblinks):
    annotation = {}
    #print(dblinks)
    for db in dblinks:
        if db == "BiGG":
            annotation["bigg.reaction"] = dblinks[db][0]
        if db == "LigandReaction":
            annotation["kegg.reaction"] = dblinks[db][0]
        if db == "MetaCyc":
            annotation["biocyc"] = dblinks[db][0]
        if db == "ModelSeedReaction":
            annotation["seed.reaction"] = dblinks[db][0]
    return annotation

def build_gene_id(str):
    if str.startswith("G_"):
        return str[2:]
    return str

def build_cpd_id(str):
    if str.startswith("M_"):
        str = str[2:]
    elif str.startswith("M-"):
        str = str[2:]
    str_fix = str
    if '-' in str_fix:
        str_fix = str_fix.replace('-', '__DASH__')
    if not str == str_fix:
        logger.debug('[Species] rename: [%s] -> [%s]', str, str_fix)
    return str

def build_rxn_id(str):
    if str.startswith("R_"):
        str = str[2:]
    elif str.startswith("R-"):
        str = str[2:]
    str_fix = str
    if '-' in str_fix:
        str_fix = str_fix.replace('-', '__DASH__')
    if not str == str_fix:
        logger.debug('[Reaction] rename: [%s] -> [%s]', str, str_fix)
    return str_fix

def get_genes(gpr):
    genes = set()
    for s in gpr:
        genes |= s
    return genes

def get_gpr_string(gpr):
    ors = []
    for ands in gpr:
        a = []
        for g in ands:
            a.append(g)
        ors.append(" and ".join(a))
    gpr_string = "(" + (") or (".join(ors)) + ")"
    if gpr_string == "()":
        return ""
    #if gpr_string.startswith("(") and gpr_string.endswith(")"):
    #    gpr_string = gpr_string[1:-1].strip()
    return gpr_string

def get_gpr(mr):
    gpr = []
    for mrp in mr['modelReactionProteins']:
        #print(mrp.keys())
        gpr_and = set()
        for mrps in mrp['modelReactionProteinSubunits']:
            #print(mrps.keys())
            for feature_ref in mrps['feature_refs']:
                gpr_and.add(get_id_from_ref(feature_ref))
        if len(gpr_and) > 0:
            gpr.append(gpr_and)
    return gpr



def read_modelseed_compound_aliases2(df):
    aliases = {}
    for a, row in df.iterrows():
        cpd_id = row[0]
        database_id = row[1]
        database = row[2]
        if not cpd_id in aliases:
            aliases[cpd_id] = {}
        
        if database == 'BiGG' or 'BiGG' in database:
            aliases[cpd_id]['bigg.metabolite'] = database_id
        if database == 'MetaCyc' or 'MetaCyc' in database:
            aliases[cpd_id]['biocyc'] = database_id
        if (database == 'KEGG' or 'KEGG' in database) and database_id[0] == 'C':
            aliases[cpd_id]['kegg.compound'] = database_id
    return aliases

def read_modelseed_reaction_aliases2(df):
    aliases = {}
    for a, row in df.iterrows():
        #print(row)
        #break
        rxn_id = row[0]
        database_id = row[1]
        database = row[2]
        if not rxn_id in aliases:
            aliases[rxn_id] = {}
        if database == 'BiGG' or 'BiGG' in database:
            aliases[rxn_id]['bigg.reaction'] = database_id
        if database == 'MetaCyc' or 'MetaCyc' in database:
            aliases[rxn_id]['biocyc'] = database_id
        if (database == 'KEGG' or 'KEGG' in database) and database_id[0] == 'R':
            aliases[rxn_id]['kegg.reaction'] = database_id
    return aliases

def read_modelseed_compound_aliases(df):
    aliases = {}
    for a, row in df.iterrows():
        if row[3] == 'BiGG':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['bigg.metabolite'] = row[2]
        if row[3] == 'MetaCyc':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['biocyc'] = row[2]
        if row[3] == 'KEGG' and row[2][0] == 'C':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['kegg.compound'] = row[2]
    return aliases

def read_modelseed_reaction_aliases(df):
    aliases = {}
    for a, row in df.iterrows():
        if row[3] == 'BiGG':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['bigg.reaction'] = row[2]
        if row[3] == 'MetaCyc':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['biocyc'] = row[2]
        if row[3] == 'KEGG' and row[2][0] == 'R':
            if not row[0] in aliases:
                aliases[row[0]] = {}
            aliases[row[0]]['kegg.reaction'] = row[2]
    return aliases

def read_modelseed_compound_structures(df):
    structures = {}
    for _, row in df.iterrows():
        #print(row[0], row[1], row[3])
        if row[1] == 'InChIKey':
            if not row[0] in structures:
                structures[row[0]] = {}
            structures[row[0]]['inchikey'] = row[3]
    return structures

def get_old_alias(alias, feature, gene_aliases):
    if alias.startswith('EcoGene:'):
        gene_aliases[feature['id']]['ecogene'] = alias.split(':')[1]
    elif alias.startswith('UniProtKB/Swiss-Prot:'):
        gene_aliases[feature['id']]['uniprot'] = alias.split(':')[1]
    elif alias.startswith('NP_'):
        gene_aliases[feature['id']]['ncbiprotein'] = alias
    elif alias.startswith('ASAP:'):
        gene_aliases[feature['id']]['asap'] = alias.split(':')[1]
    elif alias.startswith('GI:'):
        gene_aliases[feature['id']]['ncbigi'] = alias
    elif alias.startswith('GeneID:'):
        gene_aliases[feature['id']]['ncbigene'] = alias.split(':')[1]
    else:
        1

def get_db_xrefs(feature):
    aliases = {}
    if 'db_xrefs' in feature:
        for ref in feature['db_xrefs']:
            db = ref[0]
            value = ref[1]
            if db == 'EcoGene':
                aliases['ecogene'] = value
            elif db == 'GeneID':
                aliases['ncbigene'] = value
    return aliases
        
def read_genome_aliases(genome):
    gene_aliases = {}
    for feature in genome['features']:
        gene_aliases[feature['id']] = {}
        if 'aliases' in feature and type(feature['aliases']) == list:
            for alias in feature['aliases']:
                if type(alias) == list:
                    for a in alias:
                        #risk parsing the first element
                        get_old_alias(a, feature, gene_aliases)
                else:
                    get_old_alias(alias, feature, gene_aliases)
                    #print('discard', alias)
        #print(feature['aliases' in ])
        db_refs = get_db_xrefs(feature)
        gene_aliases[feature['id']].update(db_refs)
    return gene_aliases

def is_transport(r):
    lhs = set()
    rhs = set()
    for m in r.reactants:
        if 'seed.compound' in m.annotation:
            lhs.add(m.annotation['seed.compound'])
        else:
            return False
    for m in r.products:
        if 'seed.compound' in m.annotation:
            rhs.add(m.annotation['seed.compound'])
        else:
            return False
    
    if not len(lhs & rhs) == 0:
        return True
    
    return False


def is_translocation(r):
    cmps = set()
    
    for m in r.reactants:
        cmps.add(m.compartment)
    for m in r.products:
        cmps.add(m.compartment)
    
    return len(cmps) > 1

def annotate_model_with_genome(model, genome):
    gene_aliases = cobrakbase.read_genome_aliases(genome)
    for g in model.genes:
        if g.id in gene_aliases:
            g.annotation.update(gene_aliases[g.id])

def annotate_model_metabolites_with_modelseed(model, modelseed):
    for m in model.metabolites:
        seed_id = None
        if 'seed.compound' in m.annotation:
            annotation = {}
            seed_id = m.annotation['seed.compound']
            seed_cpd = modelseed.get_seed_compound(seed_id)
            if not seed_cpd == None:
                if not seed_cpd.inchi == None:
                    annotation['inchi'] = seed_cpd.inchi
                if not seed_cpd.inchikey == None:
                    annotation['inchikey'] = seed_cpd.inchikey
                alias = seed_cpd.aliases
                if 'BiGG' in alias:
                    annotation['bigg.metabolite'] = list(alias['BiGG'])
                if 'KEGG' in alias:
                    annotation['kegg.compound'] = list(alias['KEGG'])
                if 'metanetx.chemical' in alias:
                    annotation['metanetx.chemical'] = list(alias['metanetx.chemical'])
                if 'MetaCyc' in alias:
                    annotation['biocyc'] = list(map(lambda x : 'META:' + x, alias['MetaCyc']))
            m.annotation.update(annotation)
            
def annotate_model_reactions_with_modelseed(model, modelseed):
    for r in model.reactions:
        seed_id = None
        if 'seed.reaction' in r.annotation:
            annotation = {}
            seed_id = r.annotation['seed.reaction']
            seed_rxn = modelseed.get_seed_reaction(seed_id)
            
            if not seed_rxn == None:
                alias = seed_rxn.aliases
                if 'BiGG' in alias:
                    annotation['bigg.reaction'] = list(alias['BiGG'])
                if 'KEGG' in alias:
                    annotation['kegg.reaction'] = list(alias['KEGG'])
                if 'rhea' in alias:
                    annotation['rhea'] = list(alias['rhea'])
                if 'metanetx.reaction' in alias:
                    annotation['metanetx.reaction'] = list(alias['metanetx.reaction'])
                if 'MetaCyc' in alias:
                    annotation['biocyc'] = list(map(lambda x : 'META:' + x, alias['MetaCyc']))
            ec_numbers = set()
            if 'Enzyme Class' in seed_rxn.ec_numbers:
                for ec in seed_rxn.ec_numbers['Enzyme Class']:
                    if ec.startswith('EC-'):
                        ec_numbers.add(ec[3:])
                    else:
                        ec_numbers.add(ec)
            if len(ec_numbers) > 0:
                annotation['ec-code'] = list(ec_numbers)
            r.annotation.update(annotation)
            
def annotate_model_with_modelseed(model, modelseed):
    annotate_model_metabolites_with_modelseed(model, modelseed)
    annotate_model_reactions_with_modelseed(model, modelseed)
            
    for r in model.reactions:
        if cobrakbase.is_translocation(r):
            if cobrakbase.is_transport(r):
                r.annotation['sbo'] = 'SBO:0000655'
            else:
                r.annotation['sbo'] = 'SBO:0000185'
                
    kbase_sinks = ['rxn13783_c0', 'rxn13784_c0', 'rxn13782_c0']
    
    for r in model.reactions:
        #r.annotation['ec-code'] = '1.1.1.1'
        #r.annotation['metanetx.reaction'] = 'MNXR103371'
        if r.id in kbase_sinks:
            r.annotation['sbo'] = 'SBO:0000632'
        if r.id.startswith('DM_'):
            r.annotation['sbo'] = 'SBO:0000628'
            
def detect_gapfilling_constrainst(kmodel, api):
    if 'gapfillings' in kmodel and len(kmodel['gapfillings']) > 0:
        print("Pulling media from gapfilling...", kmodel['gapfillings'])
        ref = api.get_object_info_from_ref(kmodel['gapfillings'][0]['media_ref'])
        o_data = api.get_object(ref.id, ref.workspace_id)
        return convert_media(o_data)
    return None


class KBaseObjectBase:
    
    def __init__(self, json=None):
        if not json == None:
            self.data = json
        else:
            self.data = {}
            
        self.otype = None
        self.otype_version = None
        
    @property
    def id(self):
        if 'id' in self.data:
            return self.data['id']
        logger.warning('missing id')
        return None
    
    @property
    def name(self):
        if 'name' in self.data:
            return self.data['name']
        return None
    
    @property
    def kbase_type(self):
        if 'name' in self.data:
            return self.data['name']
        return None
    
    @property
    def kbase_type_version(self):
        if 'name' in self.data:
            return self.data['name']
        return None
    
    @property
    def kbase_full_type(self):
        return self.kbase_type + '-' + self.kbase_type_version


############################################################
#
# Autogenerated by the KBase type compiler -
# any changes made here will be overwritten
#
############################################################

from __future__ import print_function

import json as _json
import requests as _requests
import random as _random
import os as _os

try:
    from configparser import ConfigParser as _ConfigParser  # py 3
except ImportError:
    from ConfigParser import ConfigParser as _ConfigParser  # py 2

try:
    from urllib.parse import urlparse as _urlparse  # py3
except ImportError:
    from urlparse import urlparse as _urlparse  # py2
import time

_CT = 'content-type'
_AJ = 'application/json'
_URL_SCHEME = frozenset(['http', 'https'])


def _get_token(user_id, password, auth_svc):
    # This is bandaid helper function until we get a full
    # KBase python auth client released
    # note that currently globus usernames, and therefore kbase usernames,
    # cannot contain non-ascii characters. In python 2, quote doesn't handle
    # unicode, so if this changes this client will need to change.
    body = ('user_id=' + _requests.utils.quote(user_id) + '&password=' +
            _requests.utils.quote(password) + '&fields=token')
    ret = _requests.post(auth_svc, data=body, allow_redirects=True)
    status = ret.status_code
    if status >= 200 and status <= 299:
        tok = _json.loads(ret.text)
    elif status == 403:
        raise Exception('Authentication failed: Bad user_id/password ' +
                        'combination for user %s' % (user_id))
    else:
        raise Exception(ret.text)
    return tok['token']


def _read_inifile(file=_os.environ.get(  # @ReservedAssignment
                  'KB_DEPLOYMENT_CONFIG', _os.environ['HOME'] +
                  '/.kbase_config')):
    # Another bandaid to read in the ~/.kbase_config file if one is present
    authdata = None
    if _os.path.exists(file):
        try:
            config = _ConfigParser()
            config.read(file)
            # strip down whatever we read to only what is legit
            authdata = {x: config.get('authentication', x)
                        if config.has_option('authentication', x)
                        else None for x in ('user_id', 'token',
                                            'client_secret', 'keyfile',
                                            'keyfile_passphrase', 'password')}
        except Exception as e:
            print('Error while reading INI file {}: {}'.format(file, e))
    return authdata


class ServerError(Exception):

    def __init__(self, name, code, message, data=None, error=None):
        super(Exception, self).__init__(message)
        self.name = name
        self.code = code
        self.message = '' if message is None else message
        self.data = data or error or ''
        # data = JSON RPC 2.0, error = 1.1

    def __str__(self):
        return self.name + ': ' + str(self.code) + '. ' + self.message + \
            '\n' + self.data


class _JSONObjectEncoder(_json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, frozenset):
            return list(obj)
        return _json.JSONEncoder.default(self, obj)


class BaseClient(object):
    '''
    The KBase base client.
    Required initialization arguments (positional):
    url - the url of the the service to contact:
        For SDK methods: either the url of the callback service or the
            Narrative Job Service Wrapper.
        For SDK dynamic services: the url of the Service Wizard.
        For other services: the url of the service.
    Optional arguments (keywords in positional order):
    timeout - methods will fail if they take longer than this value in seconds.
        Default 1800.
    user_id - a KBase user name.
    password - the password corresponding to the user name.
    token - a KBase authentication token.
    ignore_authrc - if True, don't read auth configuration from
        ~/.kbase_config.
    trust_all_ssl_certificates - set to True to trust self-signed certificates.
        If you don't understand the implications, leave as the default, False.
    auth_svc - the url of the KBase authorization service.
    lookup_url - set to true when contacting KBase dynamic services.
    async_job_check_time_ms - the wait time between checking job state for
        asynchronous jobs run with the run_job method.
    '''
    def __init__(
            self, url=None, timeout=30 * 60, user_id=None,
            password=None, token=None, ignore_authrc=False,
            trust_all_ssl_certificates=False,
            auth_svc='https://kbase.us/services/authorization/Sessions/Login',
            lookup_url=False,
            async_job_check_time_ms=100,
            async_job_check_time_scale_percent=150,
            async_job_check_max_time_ms=300000):
        if url is None:
            raise ValueError('A url is required')
        scheme, _, _, _, _, _ = _urlparse(url)
        if scheme not in _URL_SCHEME:
            raise ValueError(url + " isn't a valid http url")
        self.url = url
        self.timeout = int(timeout)
        self._headers = dict()
        self.trust_all_ssl_certificates = trust_all_ssl_certificates
        self.lookup_url = lookup_url
        self.async_job_check_time = async_job_check_time_ms / 1000.0
        self.async_job_check_time_scale_percent = (
            async_job_check_time_scale_percent)
        self.async_job_check_max_time = async_job_check_max_time_ms / 1000.0
        # token overrides user_id and password
        if token is not None:
            self._headers['AUTHORIZATION'] = token
        elif user_id is not None and password is not None:
            self._headers['AUTHORIZATION'] = _get_token(
                user_id, password, auth_svc)
        elif 'KB_AUTH_TOKEN' in _os.environ:
            self._headers['AUTHORIZATION'] = _os.environ.get('KB_AUTH_TOKEN')
        elif not ignore_authrc:
            authdata = _read_inifile()
            if authdata is not None:
                if authdata.get('token') is not None:
                    self._headers['AUTHORIZATION'] = authdata['token']
                elif(authdata.get('user_id') is not None and
                        authdata.get('password') is not None):
                    self._headers['AUTHORIZATION'] = _get_token(
                        authdata['user_id'], authdata['password'], auth_svc)
        if self.timeout < 1:
            raise ValueError('Timeout value must be at least 1 second')

    def _call(self, url, method, params, context=None):
        arg_hash = {'method': method,
                    'params': params,
                    'version': '1.1',
                    'id': str(_random.random())[2:]
                    }
        if context:
            if type(context) is not dict:
                raise ValueError('context is not type dict as required.')
            arg_hash['context'] = context

        body = _json.dumps(arg_hash, cls=_JSONObjectEncoder)
        ret = _requests.post(url, data=body, headers=self._headers,
                             timeout=self.timeout,
                             verify=not self.trust_all_ssl_certificates)
        ret.encoding = 'utf-8'
        if ret.status_code == 500:
            if ret.headers.get(_CT) == _AJ:
                err = ret.json()
                if 'error' in err:
                    raise ServerError(**err['error'])
                else:
                    raise ServerError('Unknown', 0, ret.text)
            else:
                raise ServerError('Unknown', 0, ret.text)
        if not ret.ok:
            ret.raise_for_status()
        resp = ret.json()
        if 'result' not in resp:
            raise ServerError('Unknown', 0, 'An unknown server error occurred')
        if not resp['result']:
            return
        if len(resp['result']) == 1:
            return resp['result'][0]
        return resp['result']

    def _get_service_url(self, service_method, service_version):
        if not self.lookup_url:
            return self.url
        service, _ = service_method.split('.')
        service_status_ret = self._call(
            self.url, 'ServiceWizard.get_service_status',
            [{'module_name': service, 'version': service_version}])
        return service_status_ret['url']

    def _set_up_context(self, service_ver=None, context=None):
        if service_ver:
            if not context:
                context = {}
            context['service_ver'] = service_ver
        return context

    def _check_job(self, service, job_id):
        return self._call(self.url, service + '._check_job', [job_id])

    def _submit_job(self, service_method, args, service_ver=None,
                    context=None):
        context = self._set_up_context(service_ver, context)
        mod, meth = service_method.split('.')
        return self._call(self.url, mod + '._' + meth + '_submit',
                          args, context)

    def run_job(self, service_method, args, service_ver=None, context=None):
        '''
        Run a SDK method asynchronously.
        Required arguments:
        service_method - the service and method to run, e.g. myserv.mymeth.
        args - a list of arguments to the method.
        Optional arguments:
        service_ver - the version of the service to run, e.g. a git hash
            or dev/beta/release.
        context - the rpc context dict.
        '''
        mod, _ = service_method.split('.')
        job_id = self._submit_job(service_method, args, service_ver, context)
        async_job_check_time = self.async_job_check_time
        while True:
            time.sleep(async_job_check_time)
            async_job_check_time = (async_job_check_time *
                                    self.async_job_check_time_scale_percent /
                                    100.0)
            if async_job_check_time > self.async_job_check_max_time:
                async_job_check_time = self.async_job_check_max_time
            job_state = self._check_job(mod, job_id)
            if job_state['finished']:
                if not job_state['result']:
                    return
                if len(job_state['result']) == 1:
                    return job_state['result'][0]
                return job_state['result']

    def call_method(self, service_method, args, service_ver=None,
                    context=None):
        '''
        Call a standard or dynamic service synchronously.
        Required arguments:
        service_method - the service and method to run, e.g. myserv.mymeth.
        args - a list of arguments to the method.
        Optional arguments:
        service_ver - the version of the service to run, e.g. a git hash
            or dev/beta/release.
        context - the rpc context dict.
        '''
        url = self._get_service_url(service_method, service_ver)
        context = self._set_up_context(service_ver, context)
        return self._call(url, service_method, args, context)



from cobrakbase.core.kbaseobject import KBaseObjectBase
from cobrakbase.core.utils import get_id_from_ref

class KBaseBiochemMedia(KBaseObjectBase):

    def get_media_constraints(self, cmp = 'e0'):
        media = {}
        for mediacompound in self.data['mediacompounds']:
            met_id = get_id_from_ref(mediacompound['compound_ref'])
            lb = -1 * mediacompound['maxFlux']
            ub = -1 * mediacompound['minFlux']
            if not cmp == None:
                met_id += '_' + cmp
            media[met_id] = (lb, ub)
        return media


from setuptools import setup, find_packages

setup(name='cobrakbase',
      version='0.1.9',
      description='KBase interface for COBRApy',
      url='https://github.com/Fxe/cobrakbase',
      author='Filipe Liu',
      author_email='fliu@anl.gov',
      license='MIT',
      packages=find_packages(),
      install_requires=[
          "cobra >= 0.14.2"
      ],
      zip_safe=False)


import logging
from cobra.core import Gene, Metabolite, Model, Reaction
from cobra.util.solver import linear_reaction_coefficients

logger = logging.getLogger(__name__)

def get_id_from_ref(str, stok='/'):
    return str.split(stok)[-1]

def build_gene_id(str):
    if str.startswith("G_"):
        return str[2:]
    return str

def build_cpd_id(str):
    if str.startswith("M_"):
        str = str[2:]
    elif str.startswith("M-"):
        str = str[2:]
    str_fix = str
    if '-' in str_fix:
        str_fix = str_fix.replace('-', '__DASH__')
    if not str == str_fix:
        logger.debug('[Species] rename: [%s] -> [%s]', str, str_fix)
    return str

def build_rxn_id(str):
    if str.startswith("R_"):
        str = str[2:]
    elif str.startswith("R-"):
        str = str[2:]
    str_fix = str
    if '-' in str_fix:
        str_fix = str_fix.replace('-', '__DASH__')
    if not str == str_fix:
        logger.debug('[Reaction] rename: [%s] -> [%s]', str, str_fix)
    return str_fix

class KBaseFBAModelToCobraBuilder():
    
    def __init__(self, fbamodel, media_const = {}, auto_sink = ["cpd02701_c0", "cpd11416_c0", "cpd15302_c0", "cpd11416_c0"]):
        self.fbamodel = fbamodel
        self.media_const = media_const
        
        self.SBO_ANNOTATION = "sbo"
        self.auto_sink = auto_sink
        self.auto_exchange = "e0"
        
        self.metabolites = {}
        self.metabolites_remap = {} #change of ids
        self.reactions = {}
        self.genes = {}
        self.biomass_reactions = set()
        
        self.sink_compounds = set()
        self.demand_compounds = set()
        self.exchange_compounds = set()
        
        self.COBRA_0_BOUND = 0
        self.COBRA_DEFAULT_LB = -1000
        self.COBRA_DEFAULT_UB = 1000
        
        self.reaction_copy_number = {}
        self.gene_annotation = {}
        
    def with_media(self, media):
        self.media_const = media.get_media_constraints()
        return self
    
    def with_genome(self, genome):
        self.gene_annotation = genome.read_genome_aliases()
        return self
    
    def with_seed_annoation(self, modelseed):
        
        return self
        
    def add_reaction(self, cobra_reaction, copy_duplicate = True):
        if not cobra_reaction.id in self.reactions:
            self.reactions[cobra_reaction.id] = cobra_reaction
        else:
            logger.warning('duplicate reaction: %s', cobra_reaction.id)
            if copy_duplicate:
                if not cobra_reaction.id in self.reaction_copy_number:
                    self.reaction_copy_number[cobra_reaction.id] = 1
                    
                copy_id = cobra_reaction.id + "_copy" + str(self.reaction_copy_number[cobra_reaction.id])
                
                logger.warning('copy reaction: [%s] -> [%s]', cobra_reaction.id, copy_id)
                
                self.reaction_copy_number[cobra_reaction.id] += 1
                cobra_reaction.id = copy_id
                self.reactions[cobra_reaction.id] = cobra_reaction
            else:
                return None
            
        return cobra_reaction
    
    def convert_modelcompartment(self, compartment, bigg=False):
        return None
    
    def convert_modelreaction_stoichiometry(self, reaction):
        object_stoichiometry = {}
        s = reaction.stoichiometry
        for metabolite_id in s:
            if metabolite_id in self.metabolites_remap:
                object_stoichiometry[self.metabolites[self.metabolites_remap[metabolite_id]]] = s[metabolite_id]
            else:
                logger.warning('[%s] undeclared species: %s', reaction.id, metabolite_id)
        return object_stoichiometry
    
    def convert_biomass_to_reaction(self, biomass):
        mr_id = biomass['id'] + "_biomass"
        name = biomass['name']
        object_stoichiometry = {}
        for biomasscompound in biomass['biomasscompounds']:
            coefficient = biomasscompound['coefficient']
            modelcompound_ref = biomasscompound['modelcompound_ref']
            metabolite_id = get_id_from_ref(modelcompound_ref)
            
            if metabolite_id in self.metabolites_remap:
                object_stoichiometry[self.metabolites[self.metabolites_remap[metabolite_id]]] = coefficient
            else:
                logger.warning('[%s] undeclared biomass species: %s', mr_id, metabolite_id)
                
        reaction = Reaction(id=build_rxn_id(mr_id), name=name, lower_bound=self.COBRA_0_BOUND, upper_bound=self.COBRA_DEFAULT_UB)
        objective_id = reaction.id
        reaction.add_metabolites(object_stoichiometry)
        #SBO:0000629 [biomass production]
        #Biomass production, often represented 'R_BIOMASS_', is usually the optimization target reaction of constraint-based models ...
        reaction.annotation[self.SBO_ANNOTATION] = "SBO:0000629" 
        return reaction
    
    def convert_modelreaction(self, reaction, bigg=False):
        mr_id = reaction.id
        name = reaction.name
        annotation = reaction.annotation
        lower_bound, upper_bound = reaction.get_reaction_constraints()
        
        id = build_rxn_id(mr_id)
        if bigg and "bigg.reaction" in annotation:
            id = annotation["bigg.reaction"]
        
        gpr = reaction.get_gpr()
        
        cobra_reaction = Reaction(id, 
                                  name=name, 
                                  lower_bound=lower_bound, 
                                  upper_bound=upper_bound)
        cobra_reaction.annotation[self.SBO_ANNOTATION] = "SBO:0000176" #biochemical reaction
        cobra_reaction.annotation.update(annotation)
        
        if id.startswith('rxn'):
            cobra_reaction.annotation["seed.reaction"] = id.split("_")[0]
        
        cobra_reaction.add_metabolites(self.convert_modelreaction_stoichiometry(reaction))
        
        cobra_reaction.gene_reaction_rule = reaction.gene_reaction_rule
        
        for genes in gpr:
            for gene in genes:
                if not gene in self.genes:
                    self.genes[gene] = gene
        
        return cobra_reaction
    
    def convert_modelcompound(self, metabolite, bigg=False):
        formula = metabolite.formula
        name = metabolite.name
        charge = metabolite.charge
        mc_id = metabolite.id
        compartment = metabolite.compartment
        annotation = metabolite.annotation
        logger.debug('%s[%s]: %s %s (%d)', mc_id, compartment, name, formula, charge)
        
        id = build_cpd_id(mc_id)
        
        if bigg and "bigg.metabolite" in annotation:
            id = annotation["bigg.metabolite"] + "_" + compartment
                #print(id)

        met = Metabolite(id, 
                         formula=formula, 
                         name=name, 
                         charge=charge, 
                         compartment=compartment)
        
        met.annotation[self.SBO_ANNOTATION] = "SBO:0000247" #simple chemical - Simple, non-repetitive chemical entity.
        if id.startswith('cpd'):
            met.annotation["seed.compound"] = id.split("_")[0]
        met.annotation.update(annotation)
        return met
    
    def build_drain_from_metabolite_id(self, cpd_id, lower_bound, upper_bound, 
                                       prefix = 'EX_', prefix_name = 'Exchange for ', sbo = 'SBO:0000627'):
        if cpd_id in self.metabolites:
            id = prefix + cpd_id
            cobra_metabolite = self.metabolites[cpd_id]
            object_stoichiometry = {cobra_metabolite : -1}
            drain_reaction = Reaction(id=id, 
                                      name= prefix_name + cobra_metabolite.name, 
                                      lower_bound=lower_bound, 
                                      upper_bound=upper_bound)
            drain_reaction.add_metabolites(object_stoichiometry)
            #  exchange reaction - ... provide matter influx or efflux to a model, for example to replenish a 
            #metabolic network with raw materials ... 
            drain_reaction.annotation[self.SBO_ANNOTATION] = sbo
            
            return drain_reaction
        
    def build(self, model_id = None):
        if model_id == None:
            model_id = self.fbamodel.id
            
        self.metabolites = {}
        self.reactions = {}
        self.biomass_reactions = set()
        
        self.sink_compounds = set()
        self.demand_compounds = set()
        self.exchange_compounds = set()
        
        for modelcompound in self.fbamodel.metabolites:
            cobra_metabolite = self.convert_modelcompound(modelcompound)
            
            if not cobra_metabolite.id in self.metabolites:
                self.metabolites[cobra_metabolite.id] = cobra_metabolite
                self.metabolites_remap[modelcompound.id] = cobra_metabolite.id
            else:
                logger.warning('duplicate compound: %s', cobra_metabolite.id)
                
            if cobra_metabolite.id in self.auto_sink:
                logger.info('Add Sink: [%s]', cobra_metabolite.id)
                self.demand_compounds.add(cobra_metabolite.id)
            if cobra_metabolite.compartment == self.auto_exchange:
                logger.debug('Add Exchange: [%s]', cobra_metabolite.id)
                self.exchange_compounds.add(cobra_metabolite.id)
                
        for modelreaction in self.fbamodel.reactions:
            cobra_reaction = self.convert_modelreaction(modelreaction)
            self.add_reaction(cobra_reaction)
                
        for biomass in self.fbamodel.data['biomasses']:
            cobra_reaction = self.convert_biomass_to_reaction(biomass)
            if not self.add_reaction(cobra_reaction) == None:
                self.biomass_reactions.add(cobra_reaction.id)
                
        for cpd_id in self.exchange_compounds:
            lower_bound = self.COBRA_DEFAULT_LB if len(self.media_const) == 0 else self.COBRA_0_BOUND
            upper_bound = self.COBRA_DEFAULT_UB
            if cpd_id in self.media_const:
                lower_bound, upper_bound = self.media_const[cpd_id]
            drain_reaction = self.build_drain_from_metabolite_id(cpd_id, lower_bound, upper_bound)
            self.add_reaction(drain_reaction)
            logger.debug('created exchange for [%s]: %s', cpd_id, drain_reaction)
        for cpd_id in self.demand_compounds:
            drain_reaction = self.build_drain_from_metabolite_id(cpd_id, self.COBRA_0_BOUND, self.COBRA_DEFAULT_UB, 
                                                                 "DM_", "Demand for ")
            self.add_reaction(drain_reaction)
            logger.debug('created demand for [%s]: %s', cpd_id, drain_reaction)
        for cpd_id in self.sink_compounds:
            drain_reaction = self.build_drain_from_metabolite_id(cpd_id, self.COBRA_0_BOUND, self.COBRA_DEFAULT_UB, 
                                                                 "SK_", "Sink for ")
            self.add_reaction(drain_reaction)
            logger.debug('created sink for [%s]: %s', cpd_id, drain_reaction)
        #print(self.genes)
        cobra_model = Model(model_id)
        for g in self.genes:
            gene = Gene(id=build_gene_id(g), name=g)
            gene.annotation[self.SBO_ANNOTATION] = "SBO:0000243"
            if g in self.gene_annotation:
                gene.annotation.update(self.gene_annotation[g])
            cobra_model.genes.append(gene)
        cobra_model.add_metabolites(list(self.metabolites.values()))
        cobra_model.add_reactions(list(self.reactions.values()))
        
        if len(self.biomass_reactions) > 0:
            default_biomass = list(self.biomass_reactions)[0]
            logger.info('Default biomass: [%s]', default_biomass)
            cobra_model.objective = default_biomass
            linear_reaction_coefficients(cobra_model)
            
        return cobra_model